<?php

if ( ! defined( 'ABSPATH' ) ) exit; // Exit if accessed directly

if ( ! class_exists( 'WooCommerce_Putler_Connector' ) ) {
    
    class WooCommerce_Putler_Connector {
        
        private $name;

        public function __construct() {

            $this->name = (defined('PUTLER_GATEWAY') ? PUTLER_GATEWAY : 'WooCommerce');

            add_filter('putler_connector_get_order_count', array( &$this, 'get_order_count') );
            add_filter('putler_connector_get_orders', array( &$this, 'get_orders') );
            add_filter('putler_connector_sub_updated', array( &$this, 'get_sub') );

            //Flag for woo2.2+
            if (version_compare ( WOOCOMMERCE_VERSION, '2.2.0', '<' )) {
                define ( 'WPC_IS_WOO22', "false" );
            } else {
                define ( 'WPC_IS_WOO22', "true" );
            }

            if(is_admin()) {
                add_action( 'before_delete_post', array(&$this, 'delete_post'), 9, 1 );
            }

        }

        //function to validate the transaction
        public function validate_transaction($transaction) {

            $response = array();

            $response_keys = array('Date','Time','Time_Zone','Source','Name','Type','Status','Currency','Gross','Fee','Net','From_Email_Address','To_Email_Address','Transaction_ID','Counterparty_Status','Address_Status','Item_Title','Item_ID','Shipping_and_Handling_Amount','Insurance_Amount','Discount','Sales_Tax','Option_1_Name','Option_1_Value','Option_2_Name','Option_2_Value','Auction_Site','Buyer_ID','Item_URL','Closing_Date','Escrow_ID','Invoice_ID','Reference_Txn_ID','Invoice_Number','Custom_Number','Quantity','Receipt_ID','Balance','Note','Address_Line_1','Address_Line_2','Town_City','State_Province','Zip_Postal_Code','Country','Contact_Phone_Number','Subscription_ID','Raw_Data','Payment_Source','External_Trans_ID','IP_Address','Sub_Meta');


            $response_float_keys = array('Gross','Fee','Net','Shipping_and_Handling_Amount','Insurance_Amount','Discount','Sales_Tax','Balance');

            foreach ($response_keys as $key) {

                $default_value = '';

                $float_value = in_array($key, $response_float_keys);

                if( !empty($float_value) ) {
                    $default_value = '0.00';
                } else {
                    if( $key == 'Currency' ) {
                        $default_value = 'USD';
                    } elseif ( $key == 'Sub_Meta' ) {
                        $default_value = json_encode(json_decode ("{}"));
                    }
                }

                $response[$key] = (!empty($transaction[$key])) ? $transaction[$key] : $default_value;
            }

            return $response;
        }


        //function to handle storing of the transactions on deletion
        public function delete_post($id) {

            if( empty($id) ) {
                return;
            }

            global $wpdb;

            $post_type = get_post_type( $id );
            $valid_post_type = array("shop_order", "shop_subscription", "scheduled-action");

            if ( empty($post_type) || in_array($post_type, $valid_post_type) === FALSE ) {
                return;
            }

            //array to store the transaction details
            $putler_transaction = array( 'Date' => get_post_modified_time( 'm/d/Y', true, $id, false ),
                                        'Time' => get_post_modified_time( 'H:i:s', true, $id, false ),
                                        'Source' => $this->name,
                                        'Time_Zone' => 'GMT',
                                        'Type' => (($post_type != 'shop_order') ? 'Recurring Payment' : 'Shopping Cart Payment Received'),
                                        'Status' => 'Delete',
                                        'Transaction_ID' => (($post_type == 'shop_order') ? $id : ''),
                                        'Subscription_ID' => (($post_type != 'shop_order') ? $id : '') ); 


            if( $post_type == 'shop_order' ) {

                //query to find if the order is a subscription renewal order
                $query = "SELECT ID as id,
                            '' as order_type
                            FROM {$wpdb->prefix}posts
                            WHERE post_parent = ".$id."
                            UNION
                            SELECT meta_value as id,
                                'renewal' as order_type
                            FROM {$wpdb->prefix}postmeta
                            WHERE post_id = ".$id."
                                AND meta_key IN ('_subscription_renewal', '_subscription_switch')";
                $results = $wpdb->get_row( $query, 'ARRAY_A' );

                if( count($results) > 0 ) {

                    $sub_id = (!empty($results['id'])) ? $results['id'] : '';

                    if( empty($results['order_type']) ) {
                        // query to get the order item ids for subscription orders
                        $query = "SELECT oim.meta_value AS pid
                                FROM {$wpdb->prefix}woocommerce_order_items AS oi
                                    JOIN {$wpdb->prefix}woocommerce_order_itemmeta AS oim 
                                        ON (oim.order_item_id = oi.order_item_id
                                                AND oi.order_item_type = 'line_item')
                                WHERE oi.order_id = ". $id ."
                                    AND oim.meta_key = '_product_id'";

                        $results = $wpdb->get_col( $query );

                        if( count($results) > 0 ) {

                            $sub = $prod = 0;

                            foreach ($results as $pid) {
                                $product_terms = wp_get_object_terms($pid, 'product_type', array('fields' => 'slugs'));

                                if( !empty($product_terms[0]) && ($product_terms[0] == 'variable-subscription' || $product_terms[0] == 'subscription') ) {
                                    $sub++;
                                } else {
                                    $prod++;
                                }
                            }
                        }

                        if( $sub == 1 && $prod == 0 ) {
                            $putler_transaction['Subscription_ID'] = $sub_id;
                            $putler_transaction['Type'] = 'Subscription Payment Received';
                        } else if( ($sub > 1 && $prod == 0) || ($sub >= 1 && $prod > 0) ) {
                            $putler_transaction['Type'] = 'Subscription Shopping Cart Payment Received';
                        }

                    } else { //renewal order
                        $putler_transaction['Subscription_ID'] = $sub_id;
                        $putler_transaction['Type'] = 'Subscription Payment Received';
                    }
                }

            } else {
                $putler_transaction['Transaction_ID'] = wp_get_post_parent_id($id) .'_D_'.get_post_modified_time( 'G', true, $id, false ) ;
            }

            $transient_name = '';
            $data = array();

            if( !empty($_GET['delete_all']) ) { //when clicked on empty trash
                $query = "SELECT option_name,
                                option_value 
                            FROM {$wpdb->prefix}options
                            WHERE option_name LIKE '%_transient_wpc_deleted_empty_trash_%' 
                                AND (UNIX_TIMESTAMP() - IFNULL(SUBSTRING_INDEX(option_name,'_transient_wpc_deleted_empty_trash_',-1),0) ) < 300";

                $results = $wpdb->get_row($query, 'ARRAY_A');
                if( count($results) > 0 ) {

                    $transient_name = (!empty($results['option_name'])) ? substr($results['option_name'], 11)  : '';
                    $data = maybe_unserialize($results['option_value']);
                    $data[] = $putler_transaction;

                } else {
                    $transient_name = 'wpc_deleted_empty_trash_'.time();
                    $data = array($putler_transaction);
                }
            } else {
                if( !empty($putler_transaction) ) {
                    $transient_name = 'wpc_deleted_'.time();
                    $data = array($putler_transaction);
                }
            }

            if( !empty($transient_name) && !empty($data) ) {
                set_transient($transient_name, $data, (DAY_IN_SECONDS * 180));
            }
        }



        public function new_status_map($v) { return $v[0]; }

        //Function to get sub data when updated
        public function get_sub_status( $status )  {

            $sub_status = array();

            $sub_status = array('wc-active' => 'Active',
                                'wc-cancelled' => 'Canceled',
                                'wc-expired' => 'Expired',
                                'wc-pending' => 'Pending',
                                'wc-pending-cancel' => 'Pending Canceled',
                                'wc-on-hold' => 'Held',
                                'wc-suspend' => 'Suspend',
                                'wc-reactivated' => 'Reactivated',
                                'wc-created' => 'Created',
                                'wc-switched' => 'Switched', //added if subscription not properly updated to v2.0
                                'trash' => 'Delete');

            $new_status = ( !empty($sub_status[$status]) ) ? $sub_status[$status] : (( !empty($sub_status['wc-'.$status]) ) ? $sub_status['wc-'.$status] : $status);

            return $new_status;
        }

        //Function to get sub data when updated
        public function get_sub( $params )  {

            if ( empty($params) || empty($params['sub_id']) ) {
                return;
            }

            global $wpdb;

            $new_status = (!empty($params['sub_details']) && !empty($params['sub_details']['post_status']) ) ? $params['sub_details']['post_status'] : ( !empty( $params['new_status'] ) ? $params['new_status'] : '' );
            $new_status = (!empty($params['sub_details']) && $params['sub_details']['post_status'] == 'trash') ? $params['sub_details']['post_status'] : $new_status; //added for handling trashed status
            $new_status = $this->get_sub_status($new_status);
            $new_status_abbr = implode('', array_map( array($this, 'new_status_map'), explode(' ', $new_status)) );

            $response = array( 'Item_Title' => 'Unknown', 'Quantity' => 1 );

            if( (empty($params['sub_details'])) ) {
                //query to get the sub details
                $query = "SELECT date_format(post_modified_gmt,'%Y-%m-%d %T') AS modified_date,
                                post_parent AS order_id
                        FROM {$wpdb->prefix}posts
                        WHERE id = ".$params['sub_id'];
                $results = $wpdb->get_results($query, 'ARRAY_A');

                if( count($results) > 0 ) {
                    foreach ($results as $result) {
                        $response ['Date'] = date('m/d/Y', (int)strtotime($result['modified_date']));
                        $response ['Time'] = date('H:i:s', (int)strtotime($result['modified_date']));
                        $response ['Time_Zone'] = 'GMT';
                        $response ['Source'] = $this->name;
                        $response ['Type'] = 'Recurring Payment';
                        $response ['Status'] = $new_status;
                        $response ['Transaction_ID'] = $result['order_id'] .'_'. $new_status_abbr .'_'. time();
                        $response ['Subscription_ID'] = $params['sub_id'];
                    }
                }
            } else {

                $response = $transaction = array();

                $transaction ['Date'] = date('m/d/Y', (int)strtotime($params['sub_details']['post_modified_gmt']));
                $transaction ['Time'] = date('H:i:s', (int)strtotime($params['sub_details']['post_modified_gmt']));
                $transaction ['Time_Zone'] = 'GMT';
                $transaction ['Source'] = $this->name;
                $transaction ['Name'] = ( !empty($params['sub_details']['_billing_first_name']) ) ? $params['sub_details']['_billing_first_name'] : '';
                $transaction ['Name'] .= ' '. (( !empty($params['sub_details']['_billing_last_name']) ) ? $params['sub_details']['_billing_last_name'] : '');
                $transaction ['Type'] = 'Recurring Payment';
                $transaction ['Status'] = $new_status;
                $transaction ['Currency'] = ( !empty($params['sub_details']['_order_currency']) ) ? $params['sub_details']['_order_currency'] : 'USD';
                $transaction ['From_Email_Address'] = ( !empty($params['sub_details']['_billing_email']) ) ? $params['sub_details']['_billing_email'] : '';
                $transaction ['Transaction_ID'] = ( $new_status != 'Created' ) ? $params['sub_details']['post_parent'] .'_'. $new_status_abbr .'_'. strtotime($params['sub_details']['post_modified_gmt']) : $params['sub_id'];
                $transaction ['Subscription_ID'] = $params['sub_id'];

                $response = $this->validate_transaction($transaction);
                
            }
            
            $pm_cond = (!empty($params['sub_details'])) ? " AND meta_key NOT IN ('_edit_lock', '_edit_last') " : " AND meta_key IN ('_order_currency', '_billing_email', '_billing_first_name', '_billing_last_name', '_payment_method', '_billing_interval', '_billing_period', '_schedule_trial_end', '_schedule_end') ";

            $sub_details = ( !empty($params['sub_meta']) ) ? $params['sub_details'] : array();
            $sub_meta = ( !empty($params['sub_meta']) ) ? $params['sub_meta'] : array();

            if( empty($sub_details) && empty($sub_meta) ) {//for non-initial data sync

                $sub_details = array( 'created_date' => ($response ['Date'] .' '. $response ['Time'] ), 'status' => $new_status  );

                $sub_meta_keys = array();

                $sub_meta_keys['_payment_method'] =  'payment_method';
                $sub_meta_keys['_billing_interval'] =  'interval';
                $sub_meta_keys['_billing_period'] =  'period';
                $sub_meta_keys['_schedule_trial_end'] =  'trial_end';
                $sub_meta_keys['_schedule_end'] =  'sub_end';

                //query to get the sub meta details
                $query = "SELECT meta_key,
                                meta_value
                        FROM {$wpdb->prefix}postmeta
                        WHERE post_id = ".$params['sub_id']."
                            $pm_cond
                        GROUP BY post_id, meta_key";
                $results = $wpdb->get_results($query, 'ARRAY_A');

                if( count($results) > 0 ) {

                    $name = '';

                    foreach ($results as $result) {
                        if( $result['meta_key'] == '_order_currency' ) {
                            $response ['Currency'] = $result['meta_value'];
                        } else if( $result['meta_key'] == '_billing_email' ) {
                            $response ['From_Email_Address'] = $result['meta_value'];
                        } else if( $result['meta_key'] == '_billing_first_name' ) {
                            $name = $result['meta_value'];
                        } else if( $result['meta_key'] == '_billing_last_name' ) {
                            $name .= ' '. $result['meta_value'];
                        }

                        $response ['Name'] = $name;

                        //code for sub meta when sub is updated
                        if( !empty($params['sub_details']) ) {
                            $sub_details[$result['meta_key']] = $result['meta_value'];
                        }

                        if( !empty($result['meta_key']) && !empty($sub_meta_keys[$result['meta_key']]) ) {
                            $sub_meta[$sub_meta_keys[$result['meta_key']]] = $result['meta_value'];
                        }
                    }
                }
            } else {

                if( !empty($sub_details['post_modified_gmt']) ) {
                    $sub_details['created_date'] = $sub_details['post_modified_gmt'];
                    unset($sub_details['post_created_gmt']);
                    unset($sub_details['post_modified_gmt']);
                    unset($sub_details['post_parent']);
                }
            }

            $response ['Raw_Data'] = '';

            if( !empty($params['sub_details']) && empty($params['new_status']) ) {
                // $response ['Raw_Data'] = maybe_serialize(array( 'subscriptions' => array( $params['sub_id'] => $sub_details ) ));
                if( empty($params['trash']) ) {
                    $response ['Raw_Data'] = json_encode(array( 'subscriptions' => array( $params['sub_id'] => $sub_details ) ));
                }
            }

            $response ['Payment_Source'] = '';
            $response ['External_Trans_ID'] = '';
            $response ['IP_Address'] = '';

            if( !empty($sub_meta) ) {
                $sub_meta['status'] = $new_status;
                $response ['Sub_Meta'] = json_encode($sub_meta);
            } else {
                $response ['Sub_Meta'] = json_encode(json_decode ("{}"));
            }

            $params['data'] = array($response);
            return $params;
        }

        public function get_order_count( $count )  {
            global $wpdb;
            $order_count = 0;
            $post_order_cond = '';
            
            //Flag for woo2.2+
            if ( (defined('WPC_IS_WOO22') && WPC_IS_WOO22) ) {
                $post_order_cond = " AND posts.post_status NOT IN ('trash', 'auto-draft', 'draft')";
            } else {                
                $post_order_cond = " AND posts.post_status IN ('publish','draft')";
            }

            $query_to_fetch_order_count = "SELECT COUNT(posts.ID) as id
                                            FROM {$wpdb->prefix}posts AS posts 
                                            WHERE posts.post_type IN ('shop_order', 'shop_order_refund') 
                                                AND posts.post_date_gmt != '0000-00-00 00:00:00'
                                                 $post_order_cond";
            
            $order_count_result = $wpdb->get_col( $query_to_fetch_order_count );

            if( !empty( $order_count_result ) ) {
                    $order_count = $order_count_result[0];
            }
            
            return $count + $order_count;
        }

        public function get_orders( $params )  {
            global $wpdb;
            $orders = array();
            
            $wc_order_status = (defined('WPC_IS_WOO22') && WPC_IS_WOO22 == 'true') ? wc_get_order_statuses() : array();

            //Code to get the last order sent
            
            $cond = '';

            if ( empty($params['order_id']) ) {
                // $start_limit = (isset($params[ $this->name ]['start_limit'])) ? $params[ $this->name ]['start_limit'] : 0;
                $start_limit = (isset($params['offset'])) ? $params['offset'] : 0;
                $batch_limit = (isset($params['limit'])) ? $params['limit'] : 50;    
            } else {
                $start_limit = 0;
                $batch_limit = 1;

                //For Handling Refund Transactions
                if( get_post_status($params['order_id']) == 'wc-refunded' ) {
                    $query = "SELECT MAX(id) AS id
                                FROM {$wpdb->prefix}posts
                                WHERE post_type = 'shop_order_refund'
                                    AND post_parent = ".$params['order_id'];
                    $refund_id = $wpdb->get_var($query);

                    if( !empty($refund_id) ) {
                        $params ['refund_parent_id'] = $params['order_id'];
                        $params ['order_id'] = $refund_id;
                    }
                }
                
                $cond = ' AND posts.ID IN(' .intval($params['order_id']). ') '; 
            }

            //Flag for woo2.2+
            if (defined('WPC_IS_WOO22') && WPC_IS_WOO22 == 'true') {
                $terms_post_join = '';
                $terms_select = "posts.post_status AS order_status";

                if( !empty($params['trash']) ) {
                    $post_order_cond = " AND posts.post_status = 'trash'";    
                } else {
                    $post_order_cond = " AND posts.post_status NOT IN ('auto-draft', 'draft')";
                }

            } else {
                //Code to get all the term_names along with the term_taxonomy_id in an array
                $query_order_status = "SELECT terms.name as order_status,
                                        term_taxonomy.term_taxonomy_id 
                                        FROM {$wpdb->prefix}term_taxonomy AS term_taxonomy
                                        JOIN {$wpdb->prefix}terms AS terms ON (terms.term_id = term_taxonomy.term_id)
                                        WHERE taxonomy LIKE 'shop_order_status'";
                                        
                $results_order_status = $wpdb->get_results( $query_order_status, 'ARRAY_A' );

                $order_status = array();

                foreach ($results_order_status as $results_order_status1) {
                    $order_status[$results_order_status1['term_taxonomy_id']] = $results_order_status1['order_status'];
                }

                $terms_post_join = "JOIN ".$wpdb->prefix ."term_relationships AS term_relationships ON (term_relationships.object_id = posts.ID AND posts.post_status IN ('publish','trash'))";
                $terms_select = "term_relationships.term_taxonomy_id AS term_taxonomy_id";

                if( !empty($params['trash']) ) {
                    $post_order_cond = " AND posts.post_status = 'trash'";
                } else {
                    $post_order_cond = " AND posts.post_status IN ('publish','trash')";
                }
            }
            
            //Code for handling manual refunds
            if( !empty($params ['refund_parent_id']) ) {
                $cond_type = " AND posts.post_type = 'shop_order_refund' ";
            } else if( !empty($params['order_id']) ) {
                $cond_type = " AND posts.post_type = 'shop_order' ";
            } else if( empty($params['order_id']) ) {
                $cond_type = " AND posts.post_type IN ('shop_order', 'shop_order_refund') ";
            }
            
            $query_order_details = "SELECT posts.ID as id,
                                        posts.post_type as type,
                                        posts.post_parent as parent_id,
                                        posts.post_excerpt as order_note,
                                        date_format(posts.post_date_gmt,'%Y-%m-%d %T') AS date,
                                        date_format(posts.post_modified_gmt,'%Y-%m-%d %T') AS modified_date,
                                        $terms_select
                                        FROM {$wpdb->prefix}posts AS posts 
                                             $terms_post_join
                                        WHERE posts.post_date_gmt != '0000-00-00 00:00:00'
                                            AND post_modified_gmt BETWEEN '". $params['start_date'] ."' AND '". $params['end_date'] ."'
                                                $cond_type
                                                $cond
                                                $post_order_cond
                                        GROUP BY posts.ID
                                        LIMIT ". $start_limit .",". $batch_limit;                           

             $results_order_details = $wpdb->get_results( $query_order_details, 'ARRAY_A' );
             $results_order_details_count = $wpdb->num_rows;

             if ( $results_order_details_count > 0 ) {
                 
                 $order_ids = array(); 
                 $order_trash_ids = array(); 
                 $man_refund_ids = array(); //array for storing the manual refund ids
                 $recurring_order_ids = array(); //array for storing the recurring order_ids
                 
                 foreach ( $results_order_details as $results_order_detail ) {
                     $order_ids[] = (!empty($results_order_detail['type']) && $results_order_detail['type'] == 'shop_order_refund') ? $results_order_detail['parent_id'] : $results_order_detail['id'];
                     

                     //for trash order ids
                     if( !empty($results_order_detail['order_status']) && $results_order_detail['order_status'] == 'trash' ) {
                        $order_trash_ids[] = $results_order_detail['id'];
                     }

                     if( !empty($results_order_detail['type']) && $results_order_detail['type'] == 'shop_order_refund' ) {
                        $man_refund_ids[$results_order_detail['id']] = $results_order_detail['parent_id'];
                     }
                 }

                 $order_ids = (!empty($params ['refund_parent_id'])) ? array($params ['refund_parent_id']) : $order_ids; //for handling meta data for manual refunds
                 
                    //Query to get the Order_items
                    
                    $item_details = array();
                 
                    $query_cart_items = "SELECT orderitems.order_id,
                                                orderitems.order_item_name,
                                                orderitems.order_item_id,
                                                orderitems.order_item_type AS item_type,
                                                itemmeta.meta_value AS meta_value,
                                                itemmeta.meta_key AS meta_key
                                        FROM {$wpdb->prefix}woocommerce_order_items AS orderitems 
                                                JOIN {$wpdb->prefix}woocommerce_order_itemmeta AS itemmeta 
                                                    ON (orderitems.order_item_id = itemmeta.order_item_id)
                                        WHERE orderitems.order_item_type IN ('line_item', 'shipping', 'coupon')
                                            AND orderitems.order_id IN (". implode(",",$order_ids) .")
                                        GROUP BY orderitems.order_id, orderitems.order_item_id, meta_key";
                    $results_cart_items = $wpdb->get_results ( $query_cart_items, 'ARRAY_A' );
                    $results_cart_items_count = $wpdb->num_rows;

                    $variation_ids = array();
                    $product_ids = array();
                    $coupon_codes = array();
                    
                    if ( $results_cart_items_count > 0 ) {
                        
                        foreach ( $results_cart_items as $cart_item ) {
                            $order_id = $cart_item['order_id']; 
                            $order_item_id = $cart_item['order_item_id'];
                            
                            if( !isset( $item_details[$order_id] ) ){
                                $item_details[$order_id] = array();
                                $item_details[$order_id]['tot_qty'] = 0;
                                $item_details[$order_id]['cart_items'] = array();
                                $item_details[$order_id]['coupons'] = array();
                                $item_details[$order_id]['_shipping_method_title'] = '';
                            }

                            // for getting the shipping related data
                            if ( $cart_item['item_type'] == 'shipping' ) {
                                $item_details[$order_id]['_shipping_method_title'] = $cart_item['order_item_name'];
                                continue;
                            }
                            
                            // for getting the coupon related data
                            if ( $cart_item['item_type'] == 'coupon' ) {
                                if ( !isset($item_details[$order_id]['coupons'][$cart_item['order_item_name']]) ) {
                                    $item_details[$order_id]['coupons'][$cart_item['order_item_name']] = array();
                                }

                                // storing the distinct coupon codes
                                if ( array_search($cart_item['order_item_name'], $coupon_codes) === false ) {
                                    $coupon_codes[] = $cart_item['order_item_name'];
                                }

                                $item_details[$order_id]['coupons'][$cart_item['order_item_name']]['code'] = $cart_item['order_item_name'];
                                $item_details[$order_id]['coupons'][$cart_item['order_item_name']]['amt'] = ($cart_item['meta_key'] == "discount_amount" && !empty($cart_item['meta_value'])) ? $cart_item['meta_value'] : 0;
                                continue;
                            }

                            if ( !isset($item_details[$order_id]['cart_items'][$order_item_id]) ) {
                                $item_details[$order_id]['cart_items'][$order_item_id] = array();
                                $item_details[$order_id]['tot_qty']++;
                                $item_details[$order_id]['cart_items'][$order_item_id]['product_name'] = $cart_item['order_item_name'];
                            } 
                            
                            $item_details[$order_id]['cart_items'][$order_item_id][$cart_item['meta_key']] = $cart_item['meta_value'];
                    
                            if ( $cart_item['meta_key'] == "_variation_id" && !empty($cart_item['meta_value']) ) {
                                $variation_ids [] = $cart_item['meta_value'];
                                $product_ids [] = $cart_item['meta_value'];
                            }

                            if ( $cart_item['meta_key'] == "_product_id" && !empty($cart_item['meta_value']) ) {
                                $product_ids [] = $cart_item['meta_value'];
                            }
                        }  
                    }


                    // Code to get the coupon meta data
                    $coupon_data = array();

                    if ( !empty($coupon_codes) ) {
                        $query = "SELECT p.post_date as created_date,
                                    p.post_name as slug,
                                    p.post_title as code,
                                    p.post_excerpt as descrip,
                                    pm.meta_value AS meta_value,
                                    pm.meta_key AS meta_key
                                FROM {$wpdb->prefix}posts AS p
                                    JOIN {$wpdb->prefix}postmeta AS pm ON (pm.post_id = p.id 
                                                                            AND p.post_type = 'shop_coupon'
                                                                            AND pm.meta_key NOT IN ('_edit_lock', '_edit_last'))
                                WHERE p.post_name IN ('". implode("','",$coupon_codes) ."')
                                GROUP BY p.id, meta_key";
                        $results = $wpdb->get_results ( $query, 'ARRAY_A' );

                        if ( count($results) > 0 ) {
                            foreach ($results as $result) {
                                if ( empty($coupon_data[$result['slug']]) ) {
                                    $coupon_data[$result['slug']] = array();
                                    $coupon_data[$result['slug']]['code'] = $result['code'];
                                    $coupon_data[$result['slug']]['description'] = $result['descrip'];
                                    $coupon_data[$result['slug']]['created_date'] = $result['created_date'];
                                }
                                $coupon_data[$result['slug']][$result['meta_key']] = $result['meta_value'];
                            }
                        }
                    }

                    
                    $products_sku = $sub_item_ids = array();

                    if( !empty($product_ids) ) {

                        //Code to get the SKU for the products

                        $query_sku = "SELECT postmeta.post_id AS id,
                                       postmeta.meta_value AS sku
                                      FROM {$wpdb->prefix}posts AS posts
                                        JOIN {$wpdb->prefix}postmeta AS postmeta ON (posts.id = postmeta.post_id)
                                      WHERE posts.id IN (". implode(",", array_unique($product_ids)) .")
                                        AND postmeta.meta_key IN ('_sku')
                                        AND postmeta.meta_value <> ''
                                      GROUP BY posts.id";
                        $results_sku  = $wpdb->get_results ( $query_sku , 'ARRAY_A');
                        $results_sku_count = $wpdb->num_rows;

                        if ( $results_sku_count > 0 ) {
                            foreach ($results_sku as $product_sku) {
                                $products_sku [$product_sku['id']] = $product_sku['sku'];
                            }
                        }    

                        //Code to get the type for the products
                        $query = "SELECT tt.term_taxonomy_id AS taxonomy_id
                                    FROM {$wpdb->prefix}term_taxonomy AS tt
                                        JOIN {$wpdb->prefix}terms AS t
                                            ON (t.term_id = tt.term_id AND tt.taxonomy = 'product_type')
                                    WHERE t.slug IN ('subscription', 'variable-subscription')";
                        $sub_taxonomy_ids  = $wpdb->get_col ( $query );

                        if( count($sub_taxonomy_ids) > 0 ) {
                            $query = "SELECT tr.object_id AS id
                                      FROM {$wpdb->prefix}term_relationships AS tr
                                        JOIN {$wpdb->prefix}term_taxonomy AS tt 
                                            ON (tt.term_taxonomy_id = tr.term_taxonomy_id AND tt.taxonomy = 'product_type')
                                      WHERE tt.term_taxonomy_id IN (". implode(',', array_unique($sub_taxonomy_ids)) .")";
                            $sub_item_ids  = $wpdb->get_col ( $query );
                        }
                    }

                    

                    //Query to get the variation Attributes

                    if (!empty($variation_ids)) {
                        $query_variation_att = "SELECT post_id AS id,
                                                   meta_value AS meta_value,
                                                   meta_key AS meta_key
                                                FROM {$wpdb->prefix}postmeta
                                                WHERE meta_key LIKE 'attribute_%'
                                                    AND post_id IN (". implode(",",array_unique($variation_ids)) .")
                                                GROUP BY id,meta_key";

                        $results_variation_att  = $wpdb->get_results ( $query_variation_att , 'ARRAY_A');
                        $results_variation_att_count = $wpdb->num_rows;

                        if ( $results_variation_att_count > 0) {

                            $i = 0;

                            $query_terms = "SELECT terms.slug as slug, terms.name as term_name
                                          FROM {$wpdb->prefix}terms AS terms
                                            JOIN {$wpdb->prefix}postmeta AS postmeta 
                                                ON ( postmeta.meta_value = terms.slug 
                                                        AND postmeta.meta_key LIKE 'attribute_%' ) 
                                          GROUP BY terms.slug";
                            $attributes_terms = $wpdb->get_results( $query_terms, 'ARRAY_A' );

                            $attributes = array();
                            foreach ( $attributes_terms as $attributes_term ) {
                                $attributes[$attributes_term['slug']] = $attributes_term['term_name'];
                            }

                            $variations = array();
                            // Formatting of the Variations Names
                            foreach( $results_variation_att as $variation_att ){

                                $att_name = '';
                                $att_val = '';

                                // if ( strpos($variation_att['meta_key'],'custom') === false ) { 
                                //     $variation_att['meta_value'] = $attributes[$variation_att['meta_value']];
                                // }


                                if ( empty($variations[$variation_att['id']]) ) {
                                    $i = 0;
                                    $variations[$variation_att['id']] = array();
                                }

                                if ( strpos($variation_att['meta_key'],'pa') !== false ) {
                                     $att_name = substr($variation_att['meta_key'], strpos($variation_att['meta_key'],'pa')+3);
                                     $att_val = (!empty($attributes[$variation_att['meta_value']])) ? $attributes[$variation_att['meta_value']] : '';
                                // } else if ( strpos($variation_att['meta_key'],'custom') !== false ) {
                                } else {
                                    // $att_name = 'custom';
                                    $att_name = ucfirst(substr($variation_att['meta_key'], strpos($variation_att['meta_key'],'_')+1));
                                    $att_val = $variation_att['meta_value'];
                                }    

                                if ( $i > 1 ) {

                                    if ( $i == 2 ) {
                                        $variations[$variation_att['id']][0]['option1_value'] = $variations[$variation_att['id']][0]['option1_name'] . ' : ' . $variations[$variation_att['id']][0]['option1_value'] . ', '
                                                                                                . $variations[$variation_att['id']][1]['option1_name'] . ' : ' . $variations[$variation_att['id']][1]['option1_value'];
                                        unset($variations[$variation_att['id']][1]);
                                    }

                                    $variations[$variation_att['id']][0]['option1_name'] = '';
                                    $variations[$variation_att['id']][0]['option1_value'] = $variations[$variation_att['id']][0]['option1_value'] . ', ' 
                                                                                            . $att_name . ' : ' . $att_val;

                                } else {
                                    $variations[$variation_att['id']][$i]['option1_name'] = $att_name;
                                    $variations[$variation_att['id']][$i]['option1_value'] = $att_val;
                                }

                                $i++;
                            }

                        }
                    }

                    //Code for handling subscriptions meta data
                    if ( in_array( 'woocommerce-subscriptions/woocommerce-subscriptions.php', apply_filters( 'active_plugins', get_option( 'active_plugins' ) ) ) ) {

                        $sub_ids = $sub_item_ids = $sub_details = $sub_o_ids = array();

                        $ps_cond = " ('auto-draft', 'draft') ";

                        //query to get subscriptions and renewal subscription ids
                        $query = "SELECT id AS sub_id,
                                        post_parent AS order_id,
                                        '' AS meta_key
                                FROM {$wpdb->prefix}posts
                                WHERE post_type = 'shop_subscription'
                                    AND post_parent IN (". implode(",",$order_ids) .")
                                    AND post_status NOT IN $ps_cond
                                    AND post_date_gmt != '0000-00-00 00:00:00'
                                GROUP BY id
                                UNION
                                SELECT pm.meta_value AS sub_id,
                                    pm.post_id AS order_id,
                                    pm.meta_key AS meta_key
                                FROM {$wpdb->prefix}postmeta AS pm
                                    JOIN {$wpdb->prefix}posts AS p ON (pm.post_id = p.id 
                                                                        AND p.post_type = 'shop_order'
                                                                        AND p.post_parent = 0)
                                WHERE pm.meta_key IN ('_subscription_renewal', '_subscription_switch')
                                    AND pm.post_id IN (". implode(",",$order_ids) .")
                                    AND p.post_status NOT IN $ps_cond
                                    AND p.post_date_gmt != '0000-00-00 00:00:00'
                                GROUP BY pm.meta_value, pm.post_id";
                        $results = $wpdb->get_results( $query, 'ARRAY_A' );

                        $trashed_sub_o_ids = array(); //to store sub_ids whose order has been trashed

                        if ( count($results) > 0 ) {
                            foreach ($results as $result) {

                                $trashed_o_id = array_search($result['order_id'], $order_trash_ids);

                                if( $trashed_o_id !== false ) {
                                    $trashed_sub_o_ids[] = $result['sub_id'];
                                }

                                $sub_o_ids[$result['order_id']] = $result['sub_id'];

                                if( !empty($result['meta_key']) ) {
                                    $recurring_order_ids[$result['order_id']] = $result['meta_key'];
                                }

                                // storing the distinct subscription ids
                                if ( array_search($result['sub_id'], $sub_ids) === false ) {
                                    $sub_ids[] = $result['sub_id'];                                    
                                }
                            }
                        }

                        if (!empty($sub_ids)) {

                            //query to get sub data
                            $query = "SELECT id as sub_id, 
                                            post_parent as order_id,
                                            date_format(post_date_gmt,'%Y-%m-%d %T') AS date,
                                            date_format(post_modified_gmt,'%Y-%m-%d %T') AS modified_date, 
                                            post_status as status
                                    FROM {$wpdb->prefix}posts
                                    WHERE post_type = 'shop_subscription'
                                        AND id IN (". implode(",",$sub_ids) .")
                                        AND post_status NOT IN $ps_cond
                                        AND post_date_gmt != '0000-00-00 00:00:00'
                                    GROUP BY id";
                            $results = $wpdb->get_results( $query, 'ARRAY_A' );


                            if ( count($results) > 0 ) {
                                foreach ($results as $result) {

                                    $trashed_sub_id = array_search($result['sub_id'], $trashed_sub_o_ids);

                                    // if( $trashed_sub_id !== false && $result['status'] != 'trash' ) { //code to not send subscription transaction in cases when only the renewal order is trashed
                                    //     $key = array_search($result['sub_id'], $sub_ids);
                                    //     unset($sub_ids[$key]);
                                    // }

                                    if( !isset($sub_details[$result['sub_id']]) && empty($params['order_id']) ) {
                                        $sub_details[$result['sub_id']] = array('post_created_gmt' => $result['date'],
                                                                                'post_modified_gmt' => $result['modified_date'],
                                                                                'post_status' => $this->get_sub_status($result['status']),
                                                                                'post_parent' => $result['order_id']);
                                    }
                                }
                            }

                            if( !empty( $item_details ) ) {
                                foreach ($item_details as $order_id => $item_detail) {
                                    if(empty( $sub_o_ids[$order_id] )) {
                                        continue;
                                    }

                                    if (!empty($item_details[$order_id]['subscriptions'])) {
                                        $item_details[$order_id]['subscriptions'] = array();
                                    }

                                    if( !empty($sub_details[$sub_o_ids[$order_id]]) ) {
                                        $item_details[$order_id]['subscriptions'] [$sub_o_ids[$order_id]] = array( 'created_date' => $sub_details[$sub_o_ids[$order_id]]['post_created_gmt'],
                                                                                                                        'modified_date' => $sub_details[$sub_o_ids[$order_id]]['post_modified_gmt'],
                                                                                                                        'status' => $sub_details[$sub_o_ids[$order_id]]['post_status'] );


                                        if( !empty($recurring_order_ids[$order_id]) ) {
                                            $item_details[$order_id]['subscriptions'][$recurring_order_ids[$order_id]] = 1;
                                        }
                                    }

                                    //Code for handling multiple subscriptions in single order
                                    if( !empty($sub_order_ids[$order_id]) && !is_array($sub_order_ids[$order_id]) ) {
                                        $temp = $sub_order_ids[$order_id];
                                        $sub_order_ids[$order_id] = array();
                                        $sub_order_ids[$order_id][0] = $temp;
                                        $sub_order_ids[$order_id][1] = $sub_o_ids[$order_id];
                                    } else if( !empty($sub_order_ids[$order_id]) && is_array($sub_order_ids[$order_id]) ) {
                                        $sub_order_ids[$order_id][] = $sub_o_ids[$order_id];
                                    } else {
                                        $sub_order_ids[$order_id] = $sub_o_ids[$order_id];    
                                    }
                                }
                            }

                            //Code for getting the subscription meta data
                        
                            $query = "SELECT post_id as id,
                                            meta_value as meta_value,
                                            meta_key as meta_key    
                                    FROM {$wpdb->prefix}postmeta
                                    WHERE post_id IN (". implode(",",$sub_ids) .")
                                        AND meta_key NOT IN ('_edit_lock', '_edit_last')
                                    GROUP BY id,meta_key";

                            $results = $wpdb->get_results( $query, 'ARRAY_A' );

                            if ( count($results) > 0 ) {

                                foreach ($item_details as $order_id => &$item_detail) {
                                    
                                    if ( empty($item_detail['subscriptions']) ) {
                                        continue;
                                    }

                                    foreach ($results as $result) {

                                        if (isset($item_detail['subscriptions'][$result['id']])) {
                                            $item_detail['subscriptions'][$result['id']][$result['meta_key']] = $result['meta_value'];    
                                        }

                                        //code for storing sub meta for 'Recurring Payment' transaction
                                        if( !empty($sub_details[$result['id']]) && empty($params['order_id']) ) {
                                            $sub_details[$result['id']][$result['meta_key']] = $result['meta_value'];
                                        }

                                    }        
                                }


                                $sub_order_details = array();

                                //code to get the billing meta infor from orders (useful in case the sub has not recorded any billing info)
                                if( !empty($sub_o_ids) ) {

                                    $o_ids = array_keys($sub_o_ids);

                                    $query = "SELECT post_id as id,
                                                    meta_value as meta_value,
                                                    meta_key as meta_key    
                                            FROM {$wpdb->prefix}postmeta
                                            WHERE post_id IN (". implode(",",$o_ids) .")
                                                AND meta_key IN ('_billing_first_name', '_billing_last_name', '_order_currency', '_billing_email', 'Stripe Fee')
                                            GROUP BY id,meta_key";

                                    $results = $wpdb->get_results( $query, 'ARRAY_A' );

                                    if ( count($results) > 0 ) {

                                        foreach ($results as $result) {
                                            if(!isset($sub_order_details[$result['id']])) {
                                                $sub_order_details[$result['id']] = array();
                                            }

                                            $sub_order_details[$result['id']][$result['meta_key']] = $result['meta_value'];
                                        }

                                    }
                                }


                                //Code for creating 'Recurring Payment' transactions for subscriptions
                                if ( empty($params['order_id']) ) { //only for data sync                                      

                                    foreach ($sub_ids as $sub_id) {

                                        if( empty($sub_details[$sub_id]['post_status']) ) { //code for ignoring the subscriptions with blank status
                                            continue;
                                        }

                                        //Code for checking if the billing info is blank for the subs
                                        $sub_o_id = array_search($sub_id, $sub_o_ids);

                                        $sub_details[$sub_id]['_billing_first_name'] = (empty($sub_details[$sub_id]['_billing_first_name']) && !empty($sub_o_id)) ? $sub_order_details[$sub_o_id]['_billing_first_name'] : $sub_details[$sub_id]['_billing_first_name'];
                                        $sub_details[$sub_id]['_billing_last_name'] = (empty($sub_details[$sub_id]['_billing_last_name']) && !empty($sub_o_id)) ? $sub_order_details[$sub_o_id]['_billing_last_name'] : $sub_details[$sub_id]['_billing_last_name'];
                                        $sub_details[$sub_id]['_order_currency'] = (empty($sub_details[$sub_id]['_order_currency']) && !empty($sub_o_id)) ? $sub_order_details[$sub_o_id]['_order_currency'] : $sub_details[$sub_id]['_order_currency'];
                                        $sub_details[$sub_id]['_billing_email'] = (empty($sub_details[$sub_id]['_billing_email']) && !empty($sub_o_id)) ? $sub_order_details[$sub_o_id]['_billing_email'] : $sub_details[$sub_id]['_billing_email'];
                                        $sub_details[$sub_id]['Stripe Fee'] = (empty($sub_details[$sub_id]['Stripe Fee']) && !empty($sub_o_id)) ? $sub_order_details[$sub_o_id]['Stripe Fee'] : $sub_details[$sub_id]['Stripe Fee'];

                                        //Code for creating 'Recurring Payment' 'Created' transaction for subscriptions

                                        $sub_details_og = $sub_details[$sub_id];

                                        if( trim($sub_details[$sub_id]['post_status']) != 'Delete' ) { //code for sending the created transaction only in non-trashed transactions

                                            $sub_details[$sub_id]['post_status'] = 'Created';
                                            $sub_details[$sub_id]['post_modified_gmt'] = $sub_details[$sub_id]['post_created_gmt']; 

                                            $args = array( 'sub_id' => $sub_id,
                                                            'sub_details' => ((!empty($sub_details[$sub_id])) ? $sub_details[$sub_id] : array()),
                                                            'sub_meta' => array( 'payment_method' => ((!empty($sub_details[$sub_id]) && !empty($sub_details[$sub_id]['_payment_method'])) ? $sub_details[$sub_id]['_payment_method'] : ''),
                                                                                'interval' => ((!empty($sub_details[$sub_id]) && !empty($sub_details[$sub_id]['_billing_interval'])) ? $sub_details[$sub_id]['_billing_interval'] : ''),
                                                                                'period' => ((!empty($sub_details[$sub_id]) && !empty($sub_details[$sub_id]['_billing_period'])) ? $sub_details[$sub_id]['_billing_period'] : ''),
                                                                                'trial_end' => ((!empty($sub_details[$sub_id]) && !empty($sub_details[$sub_id]['_schedule_trial_end'])) ? $sub_details[$sub_id]['_schedule_trial_end'] : ''),
                                                                                'sub_end' => ((!empty($sub_details[$sub_id]) && !empty($sub_details[$sub_id]['_schedule_end'])) ? $sub_details[$sub_id]['_schedule_end'] : '')
                                                                                ) 
                                                        );


                                            $sub_trans = apply_filters( 'putler_connector_sub_updated', $args );

                                            if( !empty($sub_trans['data']) ) {
                                                $sub_transactions[] = $sub_trans['data'][0];
                                            }
                                        }


                                        //Code for creating 'Recurring Payment' current status transaction for subscriptions


                                        $sub_details[$sub_id] = $sub_details_og;

                                        if( trim($sub_details[$sub_id]['post_status']) == 'Canceled' ) {
                                            $sub_details[$sub_id]['post_modified_gmt'] = (!empty($sub_details[$sub_id]['_schedule_cancelled'])) ? $sub_details[$sub_id]['_schedule_cancelled'] : $sub_details[$sub_id]['post_modified_gmt'];
                                        }

                                        $args = array( 'sub_id' => $sub_id,
                                                        'new_status' => ((!empty($sub_details[$sub_id]) && !empty($sub_details[$sub_id]['post_status'])) ? $sub_details[$sub_id]['post_status'] : ''),
                                                        'sub_details' => ((!empty($sub_details[$sub_id])) ? $sub_details[$sub_id] : array()),
                                                        'sub_meta' => array( 'payment_method' => ((!empty($sub_details[$sub_id]) && !empty($sub_details[$sub_id]['_payment_method'])) ? $sub_details[$sub_id]['_payment_method'] : ''),
                                                                            'interval' => ((!empty($sub_details[$sub_id]) && !empty($sub_details[$sub_id]['_billing_interval'])) ? $sub_details[$sub_id]['_billing_interval'] : ''),
                                                                            'period' => ((!empty($sub_details[$sub_id]) && !empty($sub_details[$sub_id]['_billing_period'])) ? $sub_details[$sub_id]['_billing_period'] : ''),
                                                                            'trial_end' => ((!empty($sub_details[$sub_id]) && !empty($sub_details[$sub_id]['_schedule_trial_end'])) ? $sub_details[$sub_id]['_schedule_trial_end'] : ''),
                                                                            'sub_end' => ((!empty($sub_details[$sub_id]) && !empty($sub_details[$sub_id]['_schedule_end'])) ? $sub_details[$sub_id]['_schedule_end'] : '')
                                                                            ) 
                                                    );

                                        $sub_trans = apply_filters( 'putler_connector_sub_updated', $args );

                                        if( !empty($sub_trans['data']) ) {
                                            $sub_transactions[] = $sub_trans['data'][0];
                                        }

                                    }
                                }
                            }

                            // query to get the subscription order item ids
                            if (!empty($sub_order_ids)) {

                                $sub_pids = $sub_prod_ids = array();

                                $query = "SELECT oi.order_id as id,
                                            GROUP_CONCAT(oim.meta_key order by oim.meta_id SEPARATOR ' #wpc# ') AS meta_key,
                                            GROUP_CONCAT(oim.meta_value order by oim.meta_id SEPARATOR ' #wpc# ') AS meta_value
                                        FROM {$wpdb->prefix}woocommerce_order_items AS oi
                                            JOIN {$wpdb->prefix}woocommerce_order_itemmeta AS oim 
                                                ON (oim.order_item_id = oi.order_item_id
                                                        AND oi.order_item_type = 'line_item')
                                        WHERE oi.order_id IN (". implode(",",$sub_ids) .")
                                            AND oim.meta_key IN ('_product_id', '_variation_id')
                                        GROUP BY id";

                                $results = $wpdb->get_results( $query, 'ARRAY_A' );

                                if( count($results) > 0 ) {
                                    foreach( $results as $result ) {
                                        $meta_values = explode(' #wpc# ', $result['meta_value']);
                                        $meta_key = explode(' #wpc# ', $result['meta_key']);

                                        if (count($meta_values) != count($meta_key))
                                            continue;

                                        $temp = array_combine($meta_key, $meta_values);

                                        $sub_pids[] = (!empty($temp['_variation_id'])) ? $temp['_variation_id'] : $temp['_product_id'];
                                        $sub_prod_ids[$result['id']] = (!empty($temp['_variation_id'])) ? $temp['_variation_id'] : $temp['_product_id'];
                                    }

                                    // query to get the order item ids for subscription orders
                                    $query = "SELECT oi.order_id AS id,
                                                oi.order_item_id AS item_id,
                                                oim.meta_value AS pid
                                            FROM {$wpdb->prefix}woocommerce_order_items AS oi
                                                JOIN {$wpdb->prefix}woocommerce_order_itemmeta AS oim 
                                                    ON (oim.order_item_id = oi.order_item_id
                                                            AND oi.order_item_type = 'line_item')
                                            WHERE oi.order_id IN (". implode(",",array_keys($sub_order_ids)) .")
                                                AND oim.meta_value IN (". implode(",",$sub_pids) .")
                                                AND oim.meta_key IN ('_product_id', '_variation_id')";

                                    $results = $wpdb->get_results( $query, 'ARRAY_A' );

                                    if( count($results) > 0 ) {
                                        foreach( $results as $result ) {

                                            if( empty($sub_order_ids[$result['id']]) ) {
                                                continue;
                                            }

                                            if( is_array($sub_order_ids[$result['id']]) ) {
                                                foreach ($sub_order_ids[$result['id']] as $sub_id) {

                                                    if( empty($sub_prod_ids[$sub_id]) ) {
                                                        continue;
                                                    }

                                                    if( $sub_prod_ids[$sub_id] == $result['pid'] ) {
                                                        $sub_item_ids[$result['item_id']] = $sub_id;
                                                        break;
                                                    }
                                                }
                                            } else {
                                                $sub_item_ids[$result['item_id']] = $sub_order_ids[$result['id']];
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }

                    // AND meta_key IN ('_billing_first_name' , '_billing_last_name' , '_billing_email',
                    //                             '_shipping_first_name', '_shipping_last_name', '_billing_address_1', '_billing_address_2',
                    //                             '_billing_city', '_billing_state', '_billing_country','_billing_postcode',
                    //                             '_shipping_method', '_payment_method', '_order_items', '_order_total',
                    //                             '_shipping_method_title', '_payment_method','_customer_user','_billing_phone',
                    //                             '_order_shipping', '_order_discount', '_cart_discount', '_order_tax', '_order_shipping_tax', '_order_currency', 'coupons')

                    $sub_renewal_id = $sub_parent_id = $sub_meta = array();

                    if( !empty($sub_item_ids) ) {

                        //Code for generating subscription meta
                        $sub_meta_keys = array();

                        $sub_meta_keys['_payment_method'] =  'payment_method';
                        $sub_meta_keys['_billing_interval'] =  'interval';
                        $sub_meta_keys['_billing_period'] =  'period';
                        $sub_meta_keys['_schedule_trial_end'] =  'trial_end';
                        $sub_meta_keys['_schedule_end'] =  'sub_end';

                        //query to get the sub meta details
                        $query = "SELECT post_id,
                                        meta_key,
                                        meta_value
                                FROM {$wpdb->prefix}postmeta
                                WHERE post_id IN ( ".implode(',',$sub_item_ids)." )
                                    AND meta_key IN ('_payment_method', '_billing_interval', '_billing_period', '_schedule_trial_end', '_schedule_end') 
                                GROUP BY post_id, meta_key";
                        $results = $wpdb->get_results($query, 'ARRAY_A');


                        if( count($results) > 0 ) {
                            foreach ($results as $result) {
                                if( !empty($result['meta_key']) && !empty($sub_meta_keys[$result['meta_key']]) ) {

                                    if( empty($sub_meta[$result['post_id']]) ) {
                                        $sub_meta[$result['post_id']] = array();
                                    }

                                    $sub_meta[$result['post_id']][$sub_meta_keys[$result['meta_key']]] = $result['meta_value'];
                                }

                                if( !isset($sub_meta[$result['post_id']]['status']) ) {
                                    $sub_meta[$result['post_id']]['status'] = $this->get_sub_status(get_post_status($result['post_id']));
                                }
                            }    
                        }
                        

                        //code to get sub meta for switch subscriptions
                         $query = "SELECT pm.meta_value AS sub_id,
                                        MAX(pm.post_id) AS order_id
                                  FROM {$wpdb->prefix}postmeta AS pm
                                    JOIN {$wpdb->prefix}posts AS p ON (p.id = pm.post_id 
                                                                        AND p.post_type = 'shop_order' 
                                                                        AND pm.meta_key = '_subscription_switch')
                                  WHERE pm.meta_value IN ( ".implode(',',$sub_item_ids)." )
                                  GROUP BY sub_id";
                         $results = $wpdb->get_results($query, ARRAY_A);

                        

                        

                         if( count($results) > 0 ) {
                            foreach ($results as $result) {
                                $sub_renewal_id[ $result['sub_id'] ] = $result['order_id'];
                            }
                         }

                        

                         //code to get the parent ids for switch subscriptions
                         $query = "SELECT id AS sub_id,
                                        post_parent AS order_id
                                    FROM {$wpdb->prefix}posts
                                    WHERE post_type = 'shop_subscription'
                                        AND id IN ( ".implode(',',$sub_item_ids)." )
                                    GROUP BY sub_id";
                         $results = $wpdb->get_results($query, ARRAY_A);

                        

                        

                         if( count($results) > 0 ) {
                            foreach ($results as $result) {
                                $sub_parent_id[ $result['sub_id'] ] = $result['order_id'];
                            }
                         }

                        
                    }

                    //Query to get the Order Details
                    $query_order_item_details = "SELECT post_id as id,
                                            meta_value as meta_value,
                                            meta_key as meta_key	
                                            FROM {$wpdb->prefix}postmeta
                                            WHERE post_id IN (". implode(",",$order_ids) .")
                                                AND meta_key NOT IN ('_edit_lock', '_edit_last')
                                            GROUP BY id,meta_key";

                    $results_order_item_details = $wpdb->get_results( $query_order_item_details, 'ARRAY_A' );
                    $results_order_item_details_count = $wpdb->num_rows;

                    if( $results_order_item_details_count > 0 ){
                        
                        $order_items = array();
                        // Structuring the order items
                        foreach( $results_order_item_details as $detail ){
                            
                            if( !isset( $order_items[$detail['id']] ) ){
                                $order_items[$detail['id']] = array();
                            }
                            
                            $order_items[$detail['id']][$detail['meta_key']] = $detail['meta_value'];
                        }

                        //Code for Data Mapping as per Putler
                        foreach( $results_order_details as $order_detail ){

                            $sub_trans = false; //flag for handling subscription transactions
                            
                            $order_id = (!empty($params['refund_parent_id'])) ? $params['refund_parent_id'] : (($order_detail['type'] == 'shop_order_refund') ? $man_refund_ids[$order_detail['id']] : $order_detail['id']);
                            $order_total = (!empty($params['refund_parent_id']) || $order_detail['type'] == 'shop_order_refund' ) ? round ( get_post_meta( $order_detail['id'], '_refund_amount', true ), 2 ) : round ( $order_items[$order_id]['_order_total'], 2 );
                            $date_gmt  = $order_detail['date'];
                            $dateInGMT = date('m/d/Y', (int)strtotime($date_gmt));
                            $timeInGMT = date('H:i:s', (int)strtotime($date_gmt));

                            if (defined('WPC_IS_WOO22') && WPC_IS_WOO22 == 'true') {
                                $order_status_new = (!empty($order_detail['order_status'])) ? substr($order_detail['order_status'],3) : '';
                            } else {
                                $order_status_new = (!empty($order_status[$order_detail['term_taxonomy_id']])) ? $order_status[$order_detail['term_taxonomy_id']] : '';
                            }

                            if( empty($order_status_new) ) { //code for ignoring the orders with blank status
                                continue;
                            }

                            $verbose_status = '';
                            if ($order_status_new== "on-hold" || $order_status_new== "pending" || $order_status_new== "failed") {
                                    $order_status_display = ($order_status_new != "failed") ?  'Pending' : 'Failed';
                                    $verbose_status = ($order_status_new != "pending") ? $order_status_new : '';
                            } else if ($order_status_new== "completed" || $order_status_new== "processing") {
                                    $order_status_display = 'Completed';
                                    $verbose_status = ($order_status_new != "completed") ? $order_status_new : '';
                            } else if ($order_status_new== "refunded") {
                                    $order_status_display = 'Refunded';
                            } else if ($order_status_new== "cancelled") {
                                    $order_status_display = 'Cancelled';
                            } else {
                                if (defined('WPC_IS_WOO22') && WPC_IS_WOO22 == 'true') { //for handling any custom order statuses
                                    $order_status_display = (!empty($order_detail['order_status']) && !empty($wc_order_status[$order_detail['order_status']])) ? $wc_order_status[$order_detail['order_status']] : '';

                                    if( !empty( $order_detail['order_status'] ) && $order_detail['order_status'] == 'trash' ) {
                                        $order_status_display = 'Delete';
                                        $verbose_status = $order_detail['order_status'];
                                    }

                                } else {
                                    $order_status_display = $order_status_new;

                                    if( !empty($params['trash']) ) {
                                        $order_status_display = 'Delete';
                                    }
                                }
                            }

                            // $response['date_time'] = $date_gmt;
                            $response ['Date'] = $dateInGMT;
                            $response ['Time'] = $timeInGMT;
                            $response ['Time_Zone'] = 'GMT';
                            
                            $response ['Source'] = $this->name;
                            $response ['Name'] = $order_items[$order_id]['_billing_first_name'] . ' ' . $order_items[$order_id]['_billing_last_name'];
                            // $response ['Type'] = ( $status == "refunded") ? 'Refund' : 'Shopping Cart Payment Received';
                            $response ['Type'] = 'Shopping Cart Payment Received';
                            $response ['Status'] = ucfirst( $order_status_display );
                            $response ['Currency'] = $order_items[$order_id]['_order_currency'];
                            $response ['Gross'] = $order_total;
                            $response ['Fee'] = ( !empty($order_items[$order_id]['Stripe Fee']) ) ? $order_items[$order_id]['Stripe Fee'] : 0.00;
                            $response ['Net'] = $order_total;
                            $response ['From_Email_Address'] = $order_items[$order_id]['_billing_email'];
                            $response ['To_Email_Address'] = '';
                            $response ['Transaction_ID'] = $order_id;
                            $response ['Counterparty_Status'] = '';
                            $response ['Address_Status'] = '';
                            $response ['Item_Title'] = 'Shopping Cart';
                            $response ['Item_ID'] = 0; // Set to 0 for main Order Transaction row
                            $response ['Shipping_and_Handling_Amount'] = ( isset( $order_items[$order_id]['_order_shipping'] ) && (empty($params['refund_parent_id']) && $order_detail['type'] != 'shop_order_refund') ) ? round ( $order_items[$order_id]['_order_shipping'], 2 ) : 0.00;
                            $response ['Insurance_Amount'] = '';
                            $response ['Discount'] = ( isset( $order_items[$order_id]['_order_discount'] ) && (empty($params['refund_parent_id']) && $order_detail['type'] != 'shop_order_refund') ) ? round ( $order_items[$order_id]['_order_discount'], 2 ) : 0.00;
                            $response ['Discount'] += ( isset( $order_items[$order_id]['_cart_discount'] ) && (empty($params['refund_parent_id']) && $order_detail['type'] != 'shop_order_refund') ) ? round ( $order_items[$order_id]['_cart_discount'], 2 ) : 0.00;
                            $response ['Sales_Tax'] = ( isset( $order_items[$order_id]['_order_tax'] ) && (empty($params['refund_parent_id']) && $order_detail['type'] != 'shop_order_refund' ) ) ? round ( $order_items[$order_id]['_order_tax'], 2 ) : 0.00;
                            $response ['Option_1_Name'] = '';
                            $response ['Option_1_Value'] = '';
                            $response ['Option_2_Name'] = '';
                            $response ['Option_2_Value'] = '';
                            $response ['Auction_Site'] = ($verbose_status != '') ? $verbose_status : '';
                            $response ['Buyer_ID'] = '';
                            $response ['Item_URL'] = '';
                            $response ['Closing_Date'] = '';
                            $response ['Escrow_ID'] = '';
                            $response ['Invoice_ID'] = '';
                            $response ['Reference_Txn_ID'] = '';
                            $response ['Invoice_Number'] = '';
                            $response ['Custom_Number'] = '';
                            $response ['Quantity'] = (!empty($item_details[$order_id]['tot_qty'])) ? $item_details[$order_id]['tot_qty'] : '1';
                            $response ['Receipt_ID'] = '';
                            $response ['Balance'] = '';
                            $response ['Note'] = $order_detail['order_note'];
                            $response ['Address_Line_1'] = ( !empty( $order_items[$order_id]['_billing_address_1'] ) ) ? $order_items[$order_id]['_billing_address_1'] : '';
                            $response ['Address_Line_2'] = !empty( $order_items[$order_id]['_billing_address_2'] ) ? $order_items[$order_id]['_billing_address_2'] : '';
                            $response ['Town_City'] = !empty( $order_items[$order_id]['_billing_city'] ) ? $order_items[$order_id]['_billing_city'] : '' ;
                            $response ['State_Province'] = $order_items[$order_id]['_billing_state'];
                            $response ['Zip_Postal_Code'] = !empty( $order_items[$order_id]['_billing_postcode'] ) ? $order_items[$order_id]['_billing_postcode'] : '';
                            $response ['Country'] = !empty( $order_items[$order_id]['_billing_country'] ) ? $order_items[$order_id]['_billing_country'] : '';
                            $response ['Contact_Phone_Number'] = !empty( $order_items[$order_id]['_billing_phone']) ? $order_items[$order_id]['_billing_phone'] : '';
                            $response ['Subscription_ID'] = '';

                            // coupon data
                            if ( !empty($item_details[$order_id]['coupons']) ) {
                                foreach ($item_details[$order_id]['coupons'] as $key => &$value) {
                                    $coupon_data[$key]['amt'] = ( !empty($value['amt']) ) ? round ( $value['amt'], 2 ) : 0.00;
                                    $value = (!empty($coupon_data[$key])) ? $coupon_data[$key] : $value;
                                }
                            }

                            $order_items[$order_id]['coupons'] = ( !empty( $item_details[$order_id]['coupons'] ) ) ? $item_details[$order_id]['coupons'] : array();
                            $order_items[$order_id]['_shipping_method_title'] = ( !empty( $item_details[$order_id]['_shipping_method_title'] ) ) ? $item_details[$order_id]['_shipping_method_title'] : '-';
                            
                            $order_items[$order_id]['subscriptions'] = array();

                            if( !empty( $item_details[$order_id]['subscriptions'] ) ) {
                                $order_items[$order_id]['subscriptions'] = $item_details[$order_id]['subscriptions'];

                                if( !empty($sub_item_ids) ) {

                                    $sub = $prod = 0;
                                    $order_sub_ids = array();

                                    foreach( $item_details[$order_id]['cart_items'] as $key => $value ) {
                                        if( !empty($sub_item_ids[$key]) ) {

                                            //For handling renewal orders in subscription meta
                                            if( !empty($order_items[$order_id]['_subscription_renewal']) ) {
                                                $sub_meta[$sub_item_ids[$key]]['is_renewal'] = 1;
                                            } else {
                                                if( isset($sub_meta[$sub_item_ids[$key]]['is_renewal']) ) {
                                                    unset($sub_meta[$sub_item_ids[$key]]['is_renewal']);
                                                }
                                            }

                                            //For handling switched subscription orders in subscription meta
                                            if( !empty($order_items[$order_id]['_subscription_switch']) ) {
                                                $sub_meta[$sub_item_ids[$key]]['sub_switch'] = ( !empty($sub_renewal_id[$sub_item_ids[$key]]) ) ? $sub_renewal_id[$sub_item_ids[$key]] : '';
                                                $sub_meta[$sub_item_ids[$key]]['sub_switch'] = ( empty( $sub_meta[$sub_item_ids[$key]]['sub_switch'] ) && !empty($sub_parent_id[$sub_item_ids[$key]]) ) ? $sub_parent_id[$sub_item_ids[$key]] : $sub_meta[$sub_item_ids[$key]]['sub_switch'];
                                            } else {
                                                if( isset($sub_meta[$sub_item_ids[$key]]['sub_switch']) ) {
                                                    unset($sub_meta[$sub_item_ids[$key]]['sub_switch']);
                                                }
                                            }

                                            $sub++;
                                        } else {
                                            $prod++;
                                        }
                                    }



                                    //code for getting the subscription meta data [subscription reports]
                                    if( !empty($order_sub_ids) ) {

                                        //For handling switched subscription orders
                                        if( !empty($order_items[$order_id]['_subscription_switch']) ) {

                                            $query = $wpdb->prepare("SELECT MAX(order_id) 
                                                                    FROM {$wpdb->prefix}wsr_orders
                                                                    WHERE sub_id = %d
                                                                        AND order_id < %d", $sub_details[0]['sub_id'], $order->id);
                                            $switch = $wpdb->get_var($query);
                                        }
                                    }

                                    //code for handling different subscription order conditions
                                    if( $sub == 1 && $prod == 0 ) {
                                        $response ['Type'] = 'Subscription Payment Received';
                                        $sub_id = array_keys($item_details[$order_id]['subscriptions']);
                                        $response ['Subscription_ID'] = (!empty($sub_id)) ? $sub_id[0] : '';
                                        $sub_trans = true;
                                    } else if( ($sub > 1 && $prod == 0) || ($sub >= 1 && $prod > 0) ) {
                                        $response ['Type'] = 'Subscription Shopping Cart Payment Received';
                                    }
                                }
                            }

                            if( empty($params['trash']) ) {
                                $response ['Raw_Data'] = json_encode($order_items[$order_id]); 
                            } else {
                                $response ['Raw_Data'] = '';
                            }

                            //Payment Title & Transaction id -- added from v2.6
                            if( $order_status_new != "refunded" && $order_detail['type'] != 'shop_order_refund' ) {
                                $response ['Payment_Source'] = !empty( $order_items[$order_id]['_payment_method']) ? $order_items[$order_id]['_payment_method'] : '';
                                $response ['External_Trans_ID'] = !empty( $order_items[$order_id]['_transaction_id']) ? $order_items[$order_id]['_transaction_id'] : '';

                                //for older woo versions > 2.0
                                $response ['External_Trans_ID'] = ( empty($response ['External_Trans_ID']) && !empty( $order_items[$order_id]['Transaction ID'] ) ) ? $order_items[$order_id]['Transaction ID'] : $response ['External_Trans_ID'];

                                //trans id for stripe
                                $response ['External_Trans_ID'] = ( !empty( $response ['Payment_Source'] ) && ( $response ['Payment_Source'] == "stripe" ) && !empty( $order_items[$order_id]['_stripe_charge_id']) ) ? $order_items[$order_id]['_stripe_charge_id'] : $response ['External_Trans_ID'];    

                                //trans id for 2Checkout
                                $response ['External_Trans_ID'] = ( !empty( $response ['Payment_Source'] ) && ( $response ['Payment_Source'] == "2Checkout" ) && !empty( $order_items[$order_id]['two_checkout_sale_id']) ) ? $order_items[$order_id]['two_checkout_sale_id'] : $response ['External_Trans_ID'];    
                            } else {
                                $response ['Payment_Source'] = $response ['External_Trans_ID'] = '';
                            }

                            //customer ip_address
                            $response ['IP_Address'] = !empty( $order_items[$order_id]['_customer_ip_address']) ? $order_items[$order_id]['_customer_ip_address'] : '';

                            $response ['Sub_Meta'] = json_encode(json_decode ("{}")); //for subscription meta

                            $cart_items = (!empty($item_details[$order_id]['cart_items'])) ? $item_details[$order_id]['cart_items'] : array();
                            $sub = (!empty($item_details[$order_id]['subscriptions'])) ? $item_details[$order_id]['subscriptions'] : array();

                            if( (!empty($params['order_id']) && ($order_status_new == "refunded" || !empty($params ['refund_parent_id'])) ) || 
                                (empty($params['order_id']) && $order_detail['type'] == 'shop_order_refund') ) {

                                $refund_id = ''; //for handling manual refunds

                                if( !empty($params ['refund_parent_id']) || $order_detail['type'] == 'shop_order_refund' ) {
                                    $refund_id = (!empty($params ['refund_parent_id'])) ? $params['order_id'] : $order_detail['id'];
                                }

                                $date_gmt_modified = $order_detail['modified_date'];

                                $response ['Date'] = date('m/d/Y', (int)strtotime($date_gmt_modified));
                                $response ['Time'] = date('H:i:s', (int)strtotime($date_gmt_modified));

                                $response ['Type'] = 'Refund';
                                $response ['Status'] = 'Completed';
                                $response ['Auction_Site'] = '';
                                $response ['Item_Title'] = 'Unknown';

                                // $response ['Gross'] = (!empty($params ['refund_parent_id']) || $order_detail['type'] == 'shop_order_refund') ? $order_total : -$order_total;
                                // $response ['Net'] = (!empty($params ['refund_parent_id']) || $order_detail['type'] == 'shop_order_refund') ? $order_total : -$order_total;

                                $response ['Gross'] = -$order_total; //for handling partial & full refunds
                                $response ['Net'] = -$order_total;

                                $response ['Transaction_ID'] = ((!empty($params ['refund_parent_id']) || $order_detail['type'] == 'shop_order_refund') && !empty($refund_id)) ? $order_id .'_R_'. $refund_id : $order_id . '_R';
                                $response ['Reference_Txn_ID'] = $order_id;
                                $response ['Raw_Data'] = '';

                                if( empty($cart_items) || count($cart_items) != 1 ) {
                                    $transactions [] = $response;
                                    continue;    
                                }
                                
                            } else if( $sub_trans === false && empty($params['order_id']) && $order_status_new == "refunded" ) { //For sending parent transactions for refunded orders during data sync
                                $transactions [] = $response;
                            }

                            if( $sub_trans === false && ($order_status_new != "refunded" && empty($params ['refund_parent_id']) && $order_detail['type'] != 'shop_order_refund') ) { //for handling orders not having only 1 subscription
                                $transactions [] = $response;
                            }

                            //Code for line items
                            foreach( $cart_items as $item_id => $cart_item ) {

                                $order_item = array();
                                $order_item ['Type'] = ($sub_trans === false && $response ['Type'] != 'Refund') ? 'Shopping Cart Item' : $response ['Type'];
                                $order_item ['Item_Title'] = $cart_item['product_name'];
                                
                                if ( !empty($cart_item['_variation_id']) ) {
                                    $order_item ['Item_ID'] = (isset( $products_sku[$cart_item['_variation_id']] )) ? $products_sku[$cart_item['_variation_id']] : $cart_item['_variation_id'];
                                    $product_id = $cart_item['_variation_id'];
                                } else {
                                    $order_item ['Item_ID'] = (isset( $products_sku[$cart_item['_product_id']] )) ? $products_sku[$cart_item['_product_id']] : $cart_item['_product_id'];
                                    $product_id = $cart_item['_product_id'];
                                }
                                
                                $order_item ['Gross'] = ($sub_trans === false && $response ['Type'] != 'Refund') ? round ( $cart_item['_line_total'], 2 ) : $response ['Gross'];
                                $order_item ['Quantity'] = $cart_item['_qty'];

                                //code to handle special case where order_total & Discount both are zero make the line_total = 0
                                if( $response ['Discount'] == 0.00 && $response ['Gross'] == 0.00 ) {
                                    $order_item ['Gross'] = 0.00;    
                                }

                                if ( isset($variations[$product_id]) ) {
                                    if ( $variations[$product_id][0]['option1_name'] != 'attributes' ) {
                                        $order_item ['Option_1_Name'] = (!empty($variations[$product_id][0]['option1_name'])) ? $variations[$product_id][0]['option1_name'] : '';
                                        $order_item ['Option_1_Value'] = (!empty($variations[$product_id][0]['option1_value'])) ? $variations[$product_id][0]['option1_value'] : '';
                                        $order_item ['Option_2_Name'] = (!empty($variations[$product_id][1]['option1_name'])) ? $variations[$product_id][1]['option1_name'] : '';
                                        $order_item ['Option_2_Value'] = (!empty($variations[$product_id][1]['option1_value'])) ? $variations[$product_id][1]['option1_value'] : '';
                                    } else {
                                        $order_item ['Option_1_Name'] = (!empty($variations[$product_id][0]['option1_name'])) ? $variations[$product_id][0]['option1_name'] : '';
                                        $order_item ['Option_1_Value'] = (!empty($variations[$product_id][0]['option1_value'])) ? $variations[$product_id][0]['option1_value'] : '';
                                    }    
                                }

                                $order_item ['Subscription_ID'] = (!empty($sub_item_ids[$item_id])) ? $sub_item_ids[$item_id] : ''; //Assigning the subscription id
                                
                                if( empty($params['trash']) ) {
                                    $response ['Raw_Data'] = ($sub_trans === false || $response ['Type'] == 'Refund') ? '' : $response ['Raw_Data'];
                                } else {
                                    $response ['Raw_Data'] = '';
                                }

                                if( !empty($sub_item_ids[$item_id]) && empty($params['trash']) ) { //for handling subscription meta
                                    $response ['Sub_Meta'] = json_encode( ( !empty($sub_meta[$sub_item_ids[$item_id]]) ) ? $sub_meta[$sub_item_ids[$item_id]] : json_decode ("{}"));
                                } else {
                                    $response ['Sub_Meta'] = json_encode(json_decode ("{}"));
                                }
                                
                                $response = array_merge ( $response, $order_item );

                                $transactions [] = $response;
                            }
                        }
                    }

                    if ( empty($params['order_id']) ) {

                        //Code for fetching the deleted transactions
                        $query = "SELECT option_value 
                                    FROM {$wpdb->prefix}options
                                    WHERE (option_name LIKE '%_transient_wpc_deleted_empty_trash_%' 
                                        AND convert_tz(FROM_UNIXTIME(SUBSTRING_INDEX(option_name,'_transient_wpc_deleted_empty_trash_',-1)),@@session.time_zone,'+00:00') BETWEEN '". $params['start_date'] ."' AND '". $params['end_date'] ."')
                                        OR (option_name LIKE '%_transient_wpc_deleted_%' 
                                        AND convert_tz(FROM_UNIXTIME(SUBSTRING_INDEX(option_name,'_transient_wpc_deleted_',-1)),@@session.time_zone,'+00:00') BETWEEN '". $params['start_date'] ."' AND '". $params['end_date'] ."')";

                        $results = $wpdb->get_col($query);

                        if( count($results) > 0 ) {
                            foreach ($results as $result) {

                                $deleted_trans = maybe_unserialize($result);

                                foreach ($deleted_trans as $trans) {
                                    $transactions[] = $this->validate_transaction($trans);
                                }
                            }
                        }

                        $order_count = (is_array($results_order_details)) ? count($results_order_details) : 0 ;
                        $data = (!empty($transactions)) ? $transactions : array();
                        $data = (!empty($sub_transactions)) ? array_merge($data,$sub_transactions) : $data;
                        $params[ $this->name ] = array('count' => $order_count, 'last_start_limit' => $start_limit, 'data' => $data );
                    } else {
                        $params['data'] = $transactions;
                    }
             }
           return $params;
        }
    }
}