<?php
/*
 * Plugin Name: WooCommerce Putler Connector
 * Plugin URI: http://putler.com/connector/woocommerce/
 * Description: Track WooCommerce transactions data with Putler. Insightful reporting that grows your business.
 * Version: 2.9.3
 * Author: putler, storeapps
 * Author URI: https://putler.com/
 * License: GPL 3.0
*/

if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly.
}

//Hooks
register_activation_hook ( __FILE__, 'wpc_activate' );
add_action( 'plugins_loaded', 'wpc_load' );


/**
 * Registers a plugin function to be run when the plugin is activated.
 */
function wpc_activate() {
	// Redirect to WPC
	update_option( '_wpc_activation_redirect', 'pending' );
}

$wpc_notice_msg = '';

function wpc_load() {

	global $wpc_notice_msg;

    $wpc_text_domain = 'putler_connector';

	$active_plugins = (array) get_option('active_plugins', array());

	if (is_multisite()) {
		$active_plugins = array_merge($active_plugins, get_site_option('active_sitewide_plugins', array()));
	}

	add_action ( 'admin_notices', 'wpc_admin_notices' );

	if( (!in_array('easy-digital-downloads-putler-connector/edd-putler-connector.php', $active_plugins) && !array_key_exists('easy-digital-downloads-putler-connector/edd-putler-connector.php', $active_plugins))
        && (!in_array('jigoshop-putler-connector/jigoshop-putler-connector.php', $active_plugins) && !array_key_exists('jigoshop-putler-connector/jigoshop-putler-connector.php', $active_plugins))
        && (!in_array('wp-e-commerce-putler-connector/wpec-putler-connector.php', $active_plugins) && !array_key_exists('wp-e-commerce-putler-connector/wpec-putler-connector.php', $active_plugins)) ) {

			if ( in_array('woocommerce/woocommerce.php', $active_plugins) || array_key_exists('woocommerce/woocommerce.php', $active_plugins) ) {

				if( !defined('PUTLER_GATEWAY') ) {
					define('PUTLER_GATEWAY', 'WooCommerce');
				}

				if( !defined('PUTLER_GATEWAY_PREFIX') ) {
	                define('PUTLER_GATEWAY_PREFIX', 'wpc');
	            }

				include_once 'classes/class.putler-connector.php';
				$GLOBALS['putler_connector'] = Putler_Connector::getInstance();
			    include_once 'classes/class.putler-connector-woocommerce.php';
			    if ( !isset( $GLOBALS['woocommerce_putler_connector'] ) ) {
			    	$GLOBALS['woocommerce_putler_connector'] = new WooCommerce_Putler_Connector();
				}	

				add_action( 'admin_init', 'wpc_init' );
			} else {
				$wpc_notice_msg = '<div id="notice" class="error"><p>'.
                            '<b>' . __( 'WooCommerce Putler Connector', $wpc_text_domain ) . '</b> ' . __( 'add-on requires', $wpc_text_domain ) . ' <a href="https://wordpress.org/plugins/woocommerce/">' . __( 'WooCommerce', $wpc_text_domain ) . '</a> ' . __( 'plugin. Please install and activate it.', $wpc_text_domain ).
                            '</p></div>';	
			}
	} else {
		$wpc_notice_msg = '<div id="notice" class="error"><p>'.
                            __( 'Any one of the Putler Connector\'s can be active at any given time. Please <b>deactivate all the other Putler Connector\'s.</b>', $wpc_text_domain ) . '</b> ' .
                            '</p></div>';
	}
}

function wpc_admin_notices() {

    global $wpc_notice_msg;

    if( !empty($wpc_notice_msg) ) {
        echo $wpc_notice_msg;
    }

}

function wpc_init () {
	// Init admin menu for settings etc if we are in admin
	if ( is_admin() ) {

		if ( false === get_option( '_wpc_update_redirect_29' ) && 'pending' !== get_option( '_wpc_activation_redirect' ) ) {
			update_option( '_wpc_update_redirect_29', 1 ); //flag for redirecting on update
			delete_option( '_wpc_update_redirect' ); 
			update_option( '_wpc_activation_redirect', 'pending' );
		}		

		if ( false !== get_option( '_wpc_activation_redirect' ) && (current_user_can('import') === true) ) {
        	// Delete the redirect transient
	    	delete_option( '_wpc_activation_redirect' );
			wp_redirect( admin_url('tools.php?page=putler_connector&action=wpc_activate') );
			exit;
		}

	} 
}

